from sklearn.metrics import precision_score,recall_score,f1_score,precision_recall_fscore_support
import numpy as np
from scipy.stats import pearsonr

def make_multi_label_mat(true,preds,nclasses=18):
    A=[]
    B=[]
    for i in range(nclasses):
        #print(nclasses)
        A+=np.argmax(true[i],axis=1).tolist()
        B+=np.argmax(preds[i],axis=1).tolist()
    AG=np.array(A).reshape((nclasses,len(preds[0]))).T
    AP=np.array(B).reshape((nclasses,len(preds[0]))).T
    return AG,AP

def evaluate(true,preds,nclasses=18):
    AG,AP = make_multi_label_mat(true,preds,nclasses)
    def f1(a,b):
        return (2*a*b)/(a+b)
    ma=f1_score(AG,AP,average='macro')
    ma2=f1(precision_score(AG,AP,average='macro'),recall_score(AG,AP,average='macro'))
    mi=f1(precision_score(AG,AP,average='micro'),recall_score(AG,AP,average='micro'))
    return [ma,ma2,mi]

def evaluate_regression(true,preds,nclasses=18):
    AG,AP = [np.array(t).flatten() for t in true],[np.array(p).flatten() for p in preds]
    ps = [pearsonr(AG[i],AP[i])[1] for i in range(len(AP))]
    rhos = [pearsonr(AG[i],AP[i])[0] for i in range(len(AP))]
    return [np.mean(np.nan_to_num(ps)),np.mean(np.nan_to_num(rhos))]

def extensive_evaluate(true,preds,nclasses=18):
    AG,AP = make_multi_label_mat(true,preds,nclasses)
    return precision_recall_fscore_support(AG,AP)

def extensive_evaluate_regression(true,preds,nclasses=18):
    AG,AP = [np.array(t).flatten() for t in true],[np.array(p).flatten() for p in preds]
    ps = [pearsonr(AG[i],AP[i])[1] for i in range(len(AP))]
    rhos = [pearsonr(AG[i],AP[i])[0] for i in range(len(AP))]
    return [np.mean(np.nan_to_num(ps)),np.mean(np.nan_to_num(rhos)),np.nan_to_num(np.array(rhos)),np.nan_to_num(np.array(ps))]

