import os
import argparse
from numpy.random import seed
from eval_helpers import evaluate, extensive_evaluate,extensive_evaluate_regression,evaluate_regression
import config
from data_setup import DataSet
from model import ModelConstructor

parser = argparse.ArgumentParser(description="Argument Parser for Proto Labeling")
parser.add_argument("PATH_DATASET",type=str,help="datasetpath")
parser.add_argument("runid",type=str,help="runid")

args = parser.parse_args()

seed(int(args.runid))

#prepare data
ds = DataSet(args.PATH_DATASET,LENGTH=config.LENGTH,AUXREG=config.AUXREG,SOFTMAX=config.SOFTMAX,PRETRAINED=config.PRETRAINED)
ds.prepare()
NCLASSES = ds.NCLASSES
trainxwords,devxwords,testxwords,trainxpointers,devxpointers,testxpointers,trainy,devy,testy = ds.make_data()

from model import ModelConstructor


#get model instance
model = ModelConstructor(HIER=config.HIER,LENGTH=config.LENGTH,AUXREG=config.AUXREG,SOFTMAX=config.SOFTMAX,ATT=config.ATT,MARKER_1=config.MARKER_1,MARKER_2=config.MARKER_2,EMBEDDING_DIM=config.EMBEDDING_DIM,NPOINTERS=4,LSTMUNITS=config.LSTMUNITS,NCLASSES=NCLASSES).make_model()




bestdev=0.0
for epoch in range(10):
    print("EPOCH",epoch)
    
    model.fit([trainxwords,trainxpointers],[ys for ys in trainy],epochs=1,batch_size=4,verbose=2)
    xpred = model.predict([devxwords,devxpointers])
    
    if config.SOFTMAX:
        ev=evaluate(devy[:NCLASSES],xpred[:NCLASSES],NCLASSES)
    else:
        ev=evaluate_regression(devy[:NCLASSES],xpred[:NCLASSES],NCLASSES)
    
    #if score on dev better than previous epoch, save model ev[1] is macro F1
    if ev[1] > bestdev:
        print("f1/pr on dev:",ev[1],"improvement:",ev[1]-bestdev,"saveing model...")
        if config.SOFTMAX:
            print("f1/pr on dev,extensive:",extensive_evaluate(devy[:NCLASSES],xpred[:NCLASSES],NCLASSES))
        else:
            print("f1/pr on dev,extensive:",extensive_evaluate_regression(devy[:NCLASSES],xpred[:NCLASSES],NCLASSES))

        bestdev=ev[1]
        model.save('models/'+args.runid+'my_model.h5')
        xpredt = model.predict([testxwords,testxpointers])

#print test predictions from best dev epoch this is needed in final ensemble vote
print("predictions:",[x.tolist() for x in xpredt])
print("gold:",[y.tolist() for y in testy])

#print scores on test from this model
if config.SOFTMAX:
    print(extensive_evaluate(testy[:NCLASSES],xpredt[:NCLASSES],NCLASSES))
else:
    print(extensive_evaluate_regression(testy[:NCLASSES],xpredt[:NCLASSES],NCLASSES))
    
